from __future__ import annotations

import os
import sys
import signal
import threading
import multiprocessing as mp
from typing import TYPE_CHECKING, Optional

import uvloop

from .server import Server
from .utils import create_logger, load_app
from .reloader import Reloader


if TYPE_CHECKING:
    from .config import Config

uvloop.install()


class Worker:
    def __init__(self, config: "Config"):
        self.app = config.app
        self.worker = None
        self.config = config
        self.logger = create_logger(
            __name__, self.config.log_level, self.config.log_fmt
        )
        self.server: Optional[Server] = None
        (self._read_stdout_fd, self._write_stdout_fd) = os.pipe()
        (self._read_stderr_fd, self._write_stderr_fd) = os.pipe()

    @property
    def stdout_fd(self):
        return self._read_stdout_fd

    @property
    def stderr_fd(self):
        return self._read_stderr_fd

    def run(self, blocking=False):
        if self.config.workers > 1 and self.config.reload:
            raise RuntimeError(
                "Number of workers must be 1 in auto reloading mode"
            )

        if not isinstance(self.app, str) and self.config.reload:
            raise RuntimeError(
                "app must be str or factory function in auto reloading mode"
            )

        self.worker = mp.Process(
            target=self.main,
            daemon=False,
            args=(
                self._write_stdout_fd.is_integer(),
                self._write_stderr_fd.is_integer(),
            ),
        )

        self.worker.start()

        if blocking:
            if self.config.reload:
                reloader = Reloader(self, self.config)
                reloader.main()

            else:
                stop_event = threading.Event()
                try:
                    stop_event.wait()
                except KeyboardInterrupt:
                    self.worker.join(timeout=5)
                    return 1

        return 0

    def _sync_to_stdio(self, stdout_fd: int, stderr_fd: int):
        # https://man7.org/linux/man-pages/man2/dup.2.html
        os.dup2(stdout_fd, sys.stdout.fileno())
        os.dup2(stderr_fd, sys.stderr.fileno())
        sys.stdout = sys.__stdout__ = open(1, "w", buffering=1)
        sys.stderr = sys.__stderr__ = open(2, "w", buffering=1)

    def main(self, stdout_fd: int, stderr_fd: int):
        """Entrypoint where child processes start and run"""

        self._sync_to_stdio(stdout_fd, stderr_fd)

        logger = create_logger(
            __name__, self.config.log_level, self.config.log_fmt
        )

        app = load_app(self.app)
        server = Server(
            app=app,
            config=self.config,
        )

        logger.info(f"Worker {self.pid} is running")

        # when user presses Ctrl-C, SIGINT will be sent to all processes
        # (what belongs to the process group) includes children so we
        # should catch them in children at here
        try:
            server.main()
        except KeyboardInterrupt:
            server.stop()
        finally:
            self.logger.info(f"Worker {self.pid} was stopped")

    @property
    def pid(self):
        if self.worker:
            return self.worker.pid

    def reload(self):
        if self.pid:
            self.logger.info("Worker is reloading")
            os.kill(self.pid, signal.SIGINT)
            self.join()
            self.run()

    def join(self):
        if self.worker and self.worker.is_alive():
            self.worker.join(timeout=5)



from __future__ import annotations

import logging
import asyncio
import urllib.parse
from collections import deque
from typing import TYPE_CHECKING, List, Literal, Optional, Set, Tuple

import httptools

from .uhttp import HTTPScope, HttpScopeRunner, ASGIHandler

if TYPE_CHECKING:
    from .server import ServerState
    from .config import Config


NO_BODY_METHOD = {
    "GET",
    "OPTIONS",
    "HEAD",
}


class H11Protocol(asyncio.Protocol):
    def __init__(
        self,
        app,
        server_state: "ServerState",
        logger: logging.Logger,
        access_logger: logging.Logger,
        config: "Config",
        loop: Optional[asyncio.AbstractEventLoop],
    ):
        # global scope
        self.app: "ASGIHandler" = app
        self.tasks: Set[asyncio.Task] = server_state.tasks
        self.connections: Set[asyncio.Protocol] = server_state.connections
        self.loop: asyncio.AbstractEventLoop = (
            loop or asyncio.get_running_loop()
        )
        self.lifespan = server_state.lifespan
        self.server_state = server_state
        self.logger = logger
        self.access_logger = access_logger
        self.config = config

        # connection scope
        self.parser = httptools.HttpRequestParser(self)  # type: ignore
        self.parser.set_dangerous_leniencies(lenient_data_after_close=True)
        self.client: Tuple[str, int]
        self.server: Tuple[str, int]
        self.transport: asyncio.Transport
        self.pipeline: deque["HttpScopeRunner"] = deque()
        self.ready_write: asyncio.Event

        # request state
        self.url: bytes
        self.scope: "HTTPScope"
        self.scheme: Optional[Literal["https", "http"]]
        self.headers: List[Tuple[bytes, bytes]]
        self.current_runner: Optional["HttpScopeRunner"] = None

    def connection_made(self, transport: asyncio.Transport) -> None:  # type:ignore
        self.transport = transport
        self.server = transport.get_extra_info("socket").getsockname()
        self.client = transport.get_extra_info("peername")
        self.ssl = transport.get_extra_info("sslcontext")
        self.ready_write = asyncio.Event()
        self.ready_write.set()
        if self.ssl:
            self.scheme = "https"
        else:
            self.scheme = "http"

    def pause_writing(self) -> None:
        self.ready_write.clear()

    def resume_writing(self) -> None:
        self.ready_write.set()

    def connection_lost(self, exc: Exception | None) -> None:
        self.connections.discard(self)
        return super().connection_lost(exc)

    def data_received(self, data: bytes) -> None:
        self.parser.feed_data(data)

    # -------------------- for parser ------------------------
    def on_url(self, url: bytes):
        self.url += url

    def on_message_begin(self):
        self.url = b""
        self.headers = []

        self.scope = {
            "method": b"",
            "query_string": b"",
            "path": "",
            "type": "http",
            "asgi": {
                "version": "2.5",
                "spec_version": "2.0",
            },
            "raw_path": None,
            "scheme": self.scheme,
            "http_version": "1.1",
            "root_path": self.server_state.root_path,
            "headers": self.headers,
            "client": self.client,
            "server": self.server,
            "state": self.lifespan.app_state,
        }

    def on_header(self, name: bytes, value: bytes):
        self.headers.append((name, value))

    def on_headers_complete(self):
        parsed_url = httptools.parse_url(self.url)  # type: ignore
        raw_path = parsed_url.path
        path = raw_path.decode("ascii")
        if "%" in path:
            path = urllib.parse.unquote(path)

        self.scope["method"] = self.parser.get_method().decode("ascii")
        self.scope["path"] = path
        self.scope["raw_path"] = raw_path
        self.scope["query_string"] = parsed_url.query or b""

        runner = HttpScopeRunner(
            scope=self.scope,
            app=self.app,
            transport=self.transport,
            message_event=asyncio.Event(),
            message_complete=self.scope["method"] in NO_BODY_METHOD,
            on_response_complete=self.on_response_complete,
            ready_write=self.ready_write,
            config=self.config,
            access_logger=self.access_logger,
        )

        if self.current_runner:
            self.pipeline.appendleft(runner)

        else:
            self.current_runner = runner
            self.schedule_runner(runner)

    def schedule_runner(self, runner: "HttpScopeRunner"):
        task = self.loop.create_task(runner.run())
        runner.task = task
        task.add_done_callback(self.tasks.discard)
        self.tasks.add(task)

    def on_response_complete(self):
        self.current_runner = None

        if self.pipeline:
            runner = self.pipeline.pop()
            self.schedule_runner(runner)

    def on_body(self, body: bytes):
        if self.current_runner:
            self.current_runner.set_body(body)

    def on_message_complete(self):
        if self.current_runner:
            self.current_runner.message_complete = True



from __future__ import annotations

import os
import sys
import asyncio
import multiprocessing as mp
from typing import TYPE_CHECKING, List

from .utils import create_logger, to_thread
from .worker import Worker


if TYPE_CHECKING:
    from .config import Config


class Arbiter:
    def __init__(
        self,
        config: "Config",
    ):
        if asyncio.iscoroutinefunction(config.app):
            raise RuntimeError(
                "You must use str or factory function in worker mode"
            )

        self.app = config.app
        self.config = config
        self.logger = create_logger(__name__, config.log_level, config.log_fmt)
        self.stop_event = mp.Event()
        self.workers: List[Worker] = []
        self.loop = asyncio.new_event_loop()

    def main(self):
        self._validate_config()

        self.logger.debug("Arbitter is running")

        to_thread(self.loop.run_forever, daemon=True, start=True)

        for _ in range(self.config.workers):
            self._spawn_worker()

        try:
            self.stop_event.wait()
        except KeyboardInterrupt:
            ...
        finally:
            self.stop()

        self.logger.debug("Arbiter was stopped")

    def stop(self):
        self.stop_event.set()
        for worker in self.workers:
            worker.join()
        self.loop.stop()

    def _spawn_worker(self):
        worker = Worker(self.config)
        self.workers.append(worker)
        self._sync_stdio_worker(worker)
        worker.run()

    def _pipe_fd(self, out_fd, in_fd):
        os.sendfile(out_fd, in_fd, 0, 1024)

    def _sync_stdio_worker(self, worker: Worker):
        self.loop.add_reader(
            worker.stdout_fd,
            self._pipe_fd,
            sys.stdout.fileno(),
            worker.stdout_fd,
        )
        self.loop.add_reader(
            worker.stderr_fd,
            self._pipe_fd,
            sys.stderr.fileno(),
            worker.stderr_fd,
        )

    def _validate_config(self):
        if not self.config.workers:
            raise RuntimeError("Number of workers must be greater than 0")



from __future__ import annotations

import os
import socket
import asyncio
from typing import List, Set, TYPE_CHECKING

from .utils import create_logger
from .protocol import H11Protocol
from .lifespan import Lifespan


if TYPE_CHECKING:
    from .worker import Worker
    from .uhttp import ASGIHandler
    from .config import Config


class ServerState:
    def __init__(self, lifespan: "Lifespan"):
        self.connections: Set[asyncio.Protocol] = set()
        self.tasks: Set[asyncio.Task] = set()
        self.root_path = os.getcwd()
        self.lifespan: "Lifespan" = lifespan


class Server:
    def __init__(
        self,
        app: "ASGIHandler",
        config: "Config",
    ):
        self.workers: List["Worker"] = []
        self.config = config
        self.app = app
        self.server: asyncio.Server
        self.logger = create_logger(__name__, config.log_level, config.log_fmt)
        self.access_logger = create_logger(
            "uasgi.access", "INFO", config.access_log_fmt
        )
        self.lifespan = Lifespan(self.app)
        self.state = ServerState(self.lifespan)

    def main(self):
        """Entrypoint where server starts and runs"""

        if self.config.socket is None:
            raise RuntimeError("Socket must be binded before starting server")

        asyncio.run(self.run(self.config.socket))

    async def run(self, sock: socket.socket):
        loop = asyncio.get_running_loop()

        self.server = await loop.create_server(
            protocol_factory=self.create_protocol,
            sock=sock,
            ssl=self.config.get_ssl(),
            start_serving=False,
        )

        await self.startup()

        try:
            await self.server.serve_forever()
        except asyncio.CancelledError:
            ...
        finally:
            await self.shutdown()

    def create_protocol(
        self, loop: asyncio.AbstractEventLoop | None = None
    ) -> asyncio.Protocol:
        return H11Protocol(
            app=self.app,
            server_state=self.state,
            logger=self.logger,
            access_logger=self.access_logger,
            config=self.config,
            loop=loop,
        )

    async def startup(self):
        self.logger.debug("Server is starting up")
        if self.config.lifespan:
            await self.lifespan.startup()

    async def shutdown(self):
        self.logger.debug("Server is shutting down")
        if self.config.lifespan:
            await self.lifespan.shutdown()

        await self.server.wait_closed()

    def stop(self):
        self.logger.debug("Server is stopping")
        self.server.close()



from __future__ import annotations

import socket
from typing import TYPE_CHECKING, Optional, Callable

from .utils import DEFAULT_LOG_FMT


if TYPE_CHECKING:
    from ssl import SSLContext
    from .utils import LOG_LEVEL
    from .uhttp import ASGIHandler


class Config:
    def __init__(
        self,
        app: str | Callable[[], "ASGIHandler"] | "ASGIHandler",
        host: Optional[str] = None,
        port: Optional[int] = None,
        sock: Optional["socket.socket"] = None,
        backlog: Optional[int] = None,
        workers: Optional[int] = None,
        ssl_cert_file: Optional[str] = None,
        ssl_key_file: Optional[str] = None,
        ssl: Optional["SSLContext"] = None,
        log_level: Optional["LOG_LEVEL"] = None,
        lifespan: bool = False,
        access_log: bool = True,
        log_fmt: Optional[str] = None,
        access_log_fmt: Optional[str] = None,
        reload: Optional[bool] = False,
    ):
        self.app = app
        self.host = host or "127.0.0.1"
        self.port = port or 5000
        self.socket: Optional["socket.socket"] = sock
        self.backlog = backlog or 4096
        self.workers = workers or 1
        self.ssl = ssl
        self.ssl_cert_file = ssl_cert_file
        self.ssl_key_file = ssl_key_file
        self.log_level: "LOG_LEVEL" = log_level or "INFO"
        self.lifespan = lifespan
        self.access_log = access_log
        self.log_fmt = log_fmt or DEFAULT_LOG_FMT
        self.access_log_fmt = access_log_fmt or DEFAULT_LOG_FMT
        self.reload = reload or False

    def get_ssl(self):
        from .utils import create_ssl_context

        if self.ssl:
            return self.ssl

        if self.ssl_cert_file and self.ssl_key_file:
            self.ssl = create_ssl_context(
                self.ssl_cert_file, self.ssl_key_file
            )

        return self.ssl

    def create_socket(self):
        host = self.host
        port = self.port
        sock = socket.create_server(
            address=(host, port),
            family=socket.AF_INET,
            backlog=self.backlog,
            reuse_port=True,
        )

        sock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.setblocking(False)

        sock.set_inheritable(True)

        return sock

    def setup_socket(self):
        if not self.socket:
            self.socket = self.create_socket()

    def __str__(self) -> str:
        output = ""

        def fmt(value) -> str:
            if isinstance(value, str):
                return value
            return str(value)

        title = "Starting ASGI Web Server with Configuration\n"
        output += title
        output += "=" * len(title) + "\n"

        entries = {
            "Host": self.host,
            "Port": self.port,
            "Socket": self.socket,
            "Backlog": self.backlog,
            "Workers": self.workers,
            "Auto reloading": self.reload,
            "SSL Enabled": self.ssl is not None,
            "SSL Cert File": self.ssl_cert_file,
            "SSL Key File": self.ssl_key_file,
            "Log Level": self.log_level,
            "Access Log": self.access_log,
            "Access Log Format": self.access_log_fmt or DEFAULT_LOG_FMT,
            "Log Format": self.log_fmt or DEFAULT_LOG_FMT,
            "Lifespan": self.lifespan,
        }

        max_key_len = max(len(k) for k in entries)

        for key, val in entries.items():
            line = f"{key:<{max_key_len}} : {fmt(val)}\n"
            output += line

        output += "=" * len(title) + "\n"
        return output



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Callable, Optional

import uvloop

from uasgi.worker import Worker

from .utils import LOG_LEVEL
from .config import Config
from .arbiter import Arbiter


if TYPE_CHECKING:
    from .uhttp import ASGIHandler


def run(
    app: str | Callable[[], "ASGIHandler"] | "ASGIHandler",
    host: str = "127.0.0.1",
    port: int = 5000,
    backlog: Optional[int] = 1024,
    workers: Optional[int] = 1,
    ssl_cert_file: Optional[str] = None,
    ssl_key_file: Optional[str] = None,
    log_level: LOG_LEVEL = "INFO",
    access_log: bool = True,
    lifespan: bool = True,
    reload: Optional[bool] = False,
    log_fmt: Optional[str] = None,
    access_log_fmt: Optional[str] = None,
):
    uvloop.install()

    config = Config(
        app,
        host=host,
        port=port,
        backlog=backlog,
        workers=workers,
        ssl_key_file=ssl_key_file,
        ssl_cert_file=ssl_cert_file,
        log_level=log_level,
        access_log=access_log,
        lifespan=lifespan,
        access_log_fmt=access_log_fmt,
        log_fmt=log_fmt,
        reload=reload,
    )
    config.setup_socket()

    sys.stdout.write(str(config))

    if config.workers == 1:
        return Worker(
            config=config,
        ).run(blocking=True)

    Arbiter(
        config=config,
    ).main()

    return



import os
import time
import socket
import errno
import asyncio
import logging
from typing import (
    TYPE_CHECKING,
    Callable,
    Optional,
    List,
    Tuple,
    Literal,
    Iterable,
    Dict,
    TypedDict,
    Coroutine,
)

if TYPE_CHECKING:
    from .config import Config


class ASGIInfo(TypedDict):
    version: str
    spec_version: str


class ASGIScope(TypedDict):
    asgi: ASGIInfo


ASGIHandler = Callable[[ASGIScope, Callable, Callable], Coroutine]


NO_BODY_METHODS = {
    b"GET",
    b"HEAD",
    b"OPTIONS",
}


class HTTPScope(ASGIScope):
    type: Literal["http"]
    http_version: str
    method: bytes
    scheme: Optional[Literal["https", "http", None]]
    path: str
    raw_path: Optional[bytes]
    query_string: bytes
    root_path: Optional[str]
    headers: Iterable[Tuple[bytes, bytes]]
    client: Optional[Tuple[str, int]]
    server: Optional[Tuple[str, int]]
    state: Optional[Dict]


class HttpScopeRunner:
    __slots__ = (
        "scope",
        "app",
        "transport",
        "message_event",
        "on_response_complete",
        "body",
        "task",
        "more_body",
        "message_complete",
        "ready_write",
        "config",
        "access_logger",
        "status",
        "content_length",
    )

    def __init__(
        self,
        scope: "HTTPScope",
        app: "ASGIHandler",
        transport: asyncio.Transport,
        message_event: asyncio.Event,
        on_response_complete: Callable[[], None],
        message_complete: bool,
        ready_write: asyncio.Event,
        config: "Config",
        access_logger: logging.Logger,
    ) -> None:
        self.app = app
        self.transport = transport
        self.scope = scope
        self.message_event = message_event
        self.body = b""
        self.on_response_complete = on_response_complete
        self.task: asyncio.Task
        self.more_body: bool = False
        self.message_complete: bool = message_complete
        self.ready_write = ready_write
        self.config = config
        self.access_logger = access_logger
        self.status: int = 200
        self.content_length: int = 0

    def set_body(self, body: bytes):
        self.body += body
        self.message_event.set()

    def drain_body(self) -> bytes:
        drain = self.body
        self.body = b""
        self.message_event.clear()
        return drain

    async def run(self):
        start_time = time.perf_counter_ns()
        try:
            return await self.app(self.scope, self.receive, self.send)
        except asyncio.CancelledError:
            ...
        except OSError:
            ...
        except RuntimeError:
            ...
        finally:
            if not self.more_body:
                self.on_response_complete()
                if self.config.access_log:
                    response_time = (
                        time.perf_counter_ns() - start_time
                    ) / 1_000_000
                    log = '{} - {:.2f} "{} {} {}" {} {}'.format(
                        self.scope["client"][0],  # type: ignore
                        response_time,
                        self.scope["method"],
                        self.scope["path"],
                        self.scope["http_version"],
                        self.status,
                        self.content_length,
                    )
                    self.access_logger.info(log)

    async def receive(self):
        if self.scope["method"] in NO_BODY_METHODS:
            event = {
                "type": "http.request",
                "body": None,
                "more_body": not self.message_complete,
            }
            return event

        await self.message_event.wait()
        body = self.drain_body()
        event = {
            "type": "http.request",
            "body": body,
            "more_body": not self.message_complete,
        }
        return event

    async def send(self, event):
        _type = event["type"]

        if _type == "http.response.start":
            data = HttpScopeRunner.build_http_response_header(
                status=event["status"],
                http_version="1.1",
                headers=event["headers"],
            )
            self.status = event["status"]
            self.transport.write(data)

        elif _type == "http.response.body":
            body = event.get("body")
            more_body = event.get("more_body", False)
            self.more_body = more_body

            if body:
                self.transport.write(body)
                self.content_length += len(body)

        elif _type == "http.response.zerocopysend":
            file = event["file"]
            count = event.get("count")
            asyncio.create_task(self.sendfile(file, count))
            self.more_body = False

    async def sendfile(
        self,
        fd: int,
        count: Optional[int] = None,
    ):
        count = count or 512

        offset = 0
        fsize = os.fstat(fd).st_size

        s: socket.socket = self.transport.get_extra_info("socket")
        fd_out = s.fileno()

        while offset < fsize:
            remaining_bytes = fsize - offset
            count_to_send = min(count, remaining_bytes)

            if count_to_send == 0:
                break

            await self.ready_write.wait()
            try:
                bytes_to_sent = os.sendfile(fd_out, fd, offset, count_to_send)

                if bytes_to_sent == 0:
                    raise RuntimeError("Connection have been closed")
            except BlockingIOError as e:
                if e.errno != errno.EAGAIN:
                    raise
                await asyncio.sleep(0)

            else:
                offset += count_to_send

    @classmethod
    def build_http_response_header(
        cls,
        status: int,
        http_version: str,
        headers: List[Tuple[bytes, bytes]],
    ):
        phrase = STATUS_PHRASES.get(status, "Unknown")
        buffer = bytearray()
        buffer.extend(
            f"HTTP/{http_version} {status} {phrase}\r\n".encode("ascii")
        )

        for k, v in headers:
            buffer.extend(k)
            buffer.extend(b":")
            buffer.extend(v)
            buffer.extend(b"\r\n")
        buffer.extend(b"\r\n")

        return bytes(buffer)


STATUS_PHRASES = {
    # 1xx: Informational
    100: "Continue",
    101: "Switching Protocols",
    102: "Processing",
    103: "Early Hints",
    # 2xx: Success
    200: "OK",
    201: "Created",
    202: "Accepted",
    203: "Non-Authoritative Information",
    204: "No Content",
    205: "Reset Content",
    206: "Partial Content",
    207: "Multi-Status",
    208: "Already Reported",
    226: "IM Used",
    # 3xx: Redirection
    300: "Multiple Choices",
    301: "Moved Permanently",
    302: "Found",
    303: "See Other",
    304: "Not Modified",
    305: "Use Proxy",
    307: "Temporary Redirect",
    308: "Permanent Redirect",
    # 4xx: Client Error
    400: "Bad Request",
    401: "Unauthorized",
    402: "Payment Required",
    403: "Forbidden",
    404: "Not Found",
    405: "Method Not Allowed",
    406: "Not Acceptable",
    407: "Proxy Authentication Required",
    408: "Request Timeout",
    409: "Conflict",
    410: "Gone",
    411: "Length Required",
    412: "Precondition Failed",
    413: "Payload Too Large",
    414: "URI Too Long",
    415: "Unsupported Media Type",
    416: "Range Not Satisfiable",
    417: "Expectation Failed",
    418: "I'm a teapot",
    421: "Misdirected Request",
    422: "Unprocessable Entity",
    423: "Locked",
    424: "Failed Dependency",
    425: "Too Early",
    426: "Upgrade Required",
    428: "Precondition Required",
    429: "Too Many Requests",
    431: "Request Header Fields Too Large",
    451: "Unavailable For Legal Reasons",
    # 5xx: Server Error
    500: "Internal Server Error",
    501: "Not Implemented",
    502: "Bad Gateway",
    503: "Service Unavailable",
    504: "Gateway Timeout",
    505: "HTTP Version Not Supported",
    506: "Variant Also Negotiates",
    507: "Insufficient Storage",
    508: "Loop Detected",
    510: "Not Extended",
    511: "Network Authentication Required",
}



from __future__ import annotations

import sys
import ssl
import logging
import asyncio
import threading
import importlib
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Literal,
    Optional,
    cast,
)


if TYPE_CHECKING:
    from .uhttp import ASGIHandler


LOG_LEVEL = Literal["DEBUG", "INFO", "WARNING", "ERROR"]


def create_ssl_context(
    certfile_path: str, keyfile_path: str, password=None
) -> ssl.SSLContext:
    context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)

    context.load_cert_chain(
        certfile=certfile_path, keyfile=keyfile_path, password=password
    )

    context.minimum_version = ssl.TLSVersion.TLSv1_2

    context.options |= ssl.OP_NO_SSLv2
    context.options |= ssl.OP_NO_SSLv3
    context.options |= ssl.OP_NO_TLSv1
    context.options |= ssl.OP_NO_TLSv1_1
    context.options |= ssl.OP_CIPHER_SERVER_PREFERENCE

    context.set_ciphers(
        "ECDHE+AESGCM:CHACHA20:DHE+AESGCM:!RC4:!aNULL:!eNULL:!LOW:!3DES:!MD5:!EXP:!PSK:!SRP:!DSS:!CAMELLIA:!SEED"
    )

    return context


DEFAULT_LOG_FMT = "%(asctime)s %(levelname)s %(name)s: %(message)s"


def create_logger(
    name: str,
    log_level: LOG_LEVEL,
    log_fmt: Optional[str] = DEFAULT_LOG_FMT,
):
    logger = logging.getLogger(name)
    logger.setLevel(log_level)

    if not log_fmt:
        log_fmt = DEFAULT_LOG_FMT

    if log_level in {"WARNING", "ERROR"}:
        handler = logging.StreamHandler(sys.stderr)

    else:
        handler = logging.StreamHandler(sys.stdout)

    fmt = logging.Formatter(log_fmt)
    handler.setFormatter(fmt)
    logger.addHandler(handler)
    return logger


def import_string(str_import: str) -> "ASGIHandler":
    try:
        module_str, instance_str = str_import.split(":")
    except ValueError as e:
        raise ImportError(
            "Import string must be <module path>:<attribute name> format"
        ) from e

    module = importlib.import_module(module_str)
    instance = getattr(module, instance_str)

    if instance is None:
        raise ImportError(f"{instance_str} is not in {module_str}")

    return instance


def load_app(app) -> "ASGIHandler":
    if isinstance(app, str):
        return import_string(app)

    if callable(app):
        if asyncio.iscoroutinefunction(app):
            return app

        return cast("ASGIHandler", app())

    raise ImportError("Cannot load app")


def to_thread(
    func: Callable,
    args: Optional[Iterable[Any]] = None,
    kwargs: Optional[Dict] = None,
    daemon=False,
    start=True,
    *targs,
    **tkwargs,
) -> threading.Thread:
    thread = threading.Thread(
        target=func,
        args=args or (),
        kwargs=kwargs,
        daemon=daemon,
        *targs,
        **tkwargs,
    )

    if start:
        thread.start()

    return thread



from __future__ import annotations

from typing import Optional
import click

from .main import run as _run
from .utils import LOG_LEVEL


cli = click.Group(name="uasgi", help="A High-Performance ASGI Web Server")


@cli.command()
@click.argument(
    "app",
    type=str,
    # help="The ASGI application to run (e.g., 'main:app')."
)
@click.option(
    "--host",
    type=str,
    default="127.0.0.1",
    show_default=True,
    help="The host IP address to bind to.",
)
@click.option(
    "--port",
    type=int,
    default=5000,
    show_default=True,
    help="The port to listen on.",
)
@click.option(
    "--backlog",
    type=int,
    default=1024,
    show_default=True,
    help="The maximum number of pending connections.",
)
@click.option(
    "--workers",
    type=int,
    default=1,
    help="The number of worker processes to use. Defaults to 1 worker.",
)
@click.option(
    "--ssl-cert-file",
    type=click.Path(exists=True, dir_okay=False),
    help="The path to the SSL certificate file (.pem).",
)
@click.option(
    "--ssl-key-file",
    type=click.Path(exists=True, dir_okay=False),
    help="The path to the SSL key file (.pem).",
)
@click.option(
    "--log-level",
    type=click.Choice(
        ["DEBUG", "INFO", "WARNING", "ERROR"], case_sensitive=True
    ),
    default="INFO",
    show_default=True,
    help="The logging level.",
)
@click.option(
    "--access-log/--no-access-log",
    is_flag=True,
    default=True,
    show_default=True,
    help="Enable/disable access logging.",
)
@click.option(
    "--lifespan/--no-lifespan",
    is_flag=True,
    default=True,
    show_default=True,
    help="Enable/disable ASGI lifespan protocol.",
)
@click.option(
    "--reload/--no-reload",
    is_flag=True,
    default=False,
    show_default=True,
    help="Enable/disable auto reloader.",
)
def run(
    app: str,
    host: str,
    port: int,
    backlog: Optional[int],
    workers: Optional[int],
    ssl_cert_file: Optional[str],
    ssl_key_file: Optional[str],
    log_level: LOG_LEVEL,
    access_log: bool,
    lifespan: bool,
    reload: Optional[bool],
):
    try:
        _run(
            app=app,
            host=host,
            port=port,
            backlog=backlog,
            workers=workers,
            ssl_key_file=ssl_key_file,
            ssl_cert_file=ssl_cert_file,
            log_level=log_level,
            access_log=access_log,
            lifespan=lifespan,
            reload=reload,
        )
    except RuntimeError as e:
        click.echo(str(e), err=True)



from __future__ import annotations

import asyncio
from typing import Any, Dict, Literal

from .uhttp import ASGIScope


class LifespanScope(ASGIScope):
    type: Literal["lifespan"]
    state: Dict[str, Any]


class Lifespan:
    def __init__(self, app):
        self.app = app
        self.app_state = dict()
        self.event_queue = asyncio.Queue()
        self.startup_done = asyncio.Event()
        self.shutdown_done = asyncio.Event()
        self.shutdown_complete = True
        self.startup_complete = True
        self.message = None
        self.scope: LifespanScope = {
            "type": "lifespan",
            "asgi": {"version": "2.5", "spec_version": "2.0"},
            "state": self.app_state,
        }

    async def startup(self):
        await self.event_queue.put({"type": "lifespan.startup"})
        asyncio.create_task(self.main())

        await self.startup_done.wait()

        if not self.startup_complete:
            raise RuntimeError(self.message or "Lifespan startup failed")

    async def send(self, event):
        _type = event["type"]

        if _type == "lifespan.startup.complete":
            self.startup_done.set()

        elif _type == "lifespan.startup.failed":
            self.startup_done.set()
            self.startup_complete = False
            self.message = event["message"]

        elif _type == "lifespan.shutdown.complete":
            self.shutdown_done.set()

        elif _type == "lifespan.shutdown.failed":
            self.shutdown_done.set()
            self.shutdown_complete = False
            self.message = event["message"]
        else:
            raise RuntimeError(f"Event {_type} is invalid")

    async def receive(self):
        return await self.event_queue.get()

    async def main(self):
        try:
            await self.app(self.scope, self.receive, self.send)

        except Exception:
            ...

    async def shutdown(self):
        await self.event_queue.put({"type": "lifespan.shutdown"})
        await self.shutdown_done.wait()

        if not self.shutdown_complete:
            raise RuntimeError(self.message or "Lifespan shutdown failed")



import os
import time
import threading
from typing import TYPE_CHECKING

from watchdog.events import (
    FileSystemEvent,
    FileSystemEventHandler,
    DirCreatedEvent,
    FileCreatedEvent,
    DirModifiedEvent,
    FileModifiedEvent,
    DirMovedEvent,
    FileMovedEvent,
    DirDeletedEvent,
    FileDeletedEvent,
)
from watchdog.observers import Observer

from .utils import create_logger


if TYPE_CHECKING:
    from .config import Config
    from .worker import Worker


class Reloader(FileSystemEventHandler):
    CHANGED_EVENT_TYPES = [
        DirCreatedEvent,
        DirModifiedEvent,
        DirMovedEvent,
        DirDeletedEvent,
        FileCreatedEvent,
        FileModifiedEvent,
        FileMovedEvent,
        FileDeletedEvent,
    ]

    DELAY_TO_LAST_TIME_RELOAD = 1

    def __init__(
        self,
        worker: "Worker",
        config: "Config",
    ):
        self.worker = worker
        self.config = config

        self.logger = create_logger(__name__, config.log_level, config.log_fmt)
        self.changed_event = threading.Event()
        self.stop_event = threading.Event()
        self.observer = Observer()
        self.reload_last_time = time.time()

        self.worker: "Worker"

    def on_any_event(self, event: FileSystemEvent) -> None:
        if self.should_reload(event):
            filename = event.src_path
            self.logger.info(f"{filename} changed")
            self.changed_event.set()

        return super().on_any_event(event)

    def should_reload(self, event: FileSystemEvent):
        diff = time.time() - self.reload_last_time
        if self.changed_event.is_set():
            return False

        if diff < Reloader.DELAY_TO_LAST_TIME_RELOAD:
            return False

        if event.is_directory:
            return False

        filename = str(os.path.basename(event.src_path))
        if filename.endswith(".py"):
            return True

        return False

    def reload_server(self):
        self.worker.reload()
        self.reload_last_time = time.time()

    def main(self):
        self.logger.debug("Reloader is running")
        cwd = os.getcwd()
        self.observer.schedule(
            event_handler=self,
            path=cwd,
            recursive=True,
            event_filter=Reloader.CHANGED_EVENT_TYPES,
        )
        self.observer.start()

        while not self.stop_event.is_set():
            try:
                self.changed_event.wait()

                self.reload_server()
                self.changed_event.clear()
            except KeyboardInterrupt:
                self.stop()

        self.logger.debug("Reloader is stopped")

    def stop(self):
        self.worker.join()
        self.stop_event.set()
